@@ -1,5 +1,13 @@
 /*
- * type2.c        2.2 4/30/95
+ * type2.c        2.5 9/20/95
+ *	Remailer choice of 0 causes mixmaster to choose a remailer
+ *	at random from type2.list
+ *
+ * type2.c        2.4 9/10/95
+ *	Change sendmail calls to use To: in body, not command line.
+ *
+ *
+ * type2.c        2.3 5/10/95
  *
  *      (c) Copyright 1995 by Lance Cottrell. All right reserved.
  *      The author assumes no liability for damages resulting from the
@@ -212,12 +220,13 @@ int merge_packets(BUFFER *body, byte *messageID, byte packet, byte numpackets)
 	    if((out = tempfile(foo))==NULL) return(-3);
 
 
-            /* put the destination list on the first line */
+            /* put the destination list in the file one per line */
+            fwrite(final_hop,1,strlen(final_hop),out);
             for(i=0;i<numdest;i++) {
               fwrite(destination[i],strlen(destination[i]),1,out);
-              fwrite(" ",1,1,out);
+              fwrite("\n",1,1,out);
             }
-            fwrite("\n",1,1,out);
+            fwrite("END\n",1,4,out);
 
             /* insert disclaimers */
             buff = new_buffer();
@@ -428,6 +437,7 @@ fprintf(stderr,"Length of digest = %d\n",len); /* debug */
 	R_DigestInit(&digest_context,DA_MD5);
 	R_DigestUpdate(&digest_context,b1->message,len);
 	R_DigestFinal(&digest_context,new_digest,&len);
+	add_to_random(new_digest,len);
 
 	ID = digest;
 #ifdef DEBUGGING
@@ -479,6 +489,7 @@ fprintf(stderr,"Length of digest = %d\n",len); /* debug */
 	add_to_buffer(b1,line,len);
 	R_OpenFinal(&rsa_context,line,&len);
 	add_to_buffer(b1,line,len);
+	add_to_random(b1->message,b1->length);
 
 	free_buffer(headers[0]);
 
@@ -500,6 +511,7 @@ printf("intermed hop\n"); /* debug */
 	  }
 	  memcpy(address,byteptr,80);
 	  crypt_in_buffer(innerkey,ivarray[18],body,0); /* decrypt back into the buffer */
+	  add_to_random(body->message,body->length);
 	  for(i=1; i<=20; i++) {
 	    crypt_in_buffer(innerkey,ivarray[i-1],headers[i],0); /* decrypt back into the buffer */
 	  }
@@ -509,7 +521,7 @@ printf("intermed hop\n"); /* debug */
 	  fprintf(fptr,"temp");  /* make sure the file is created */
 	  fclose(fptr);
 	  if(check_packetID(packetID)){
-	    if(send_new_packet(headers,body,address,line)<0) return(-1);
+	    if(send_new_packet(headers,body,address,line,0)<0) return(-1);
 	  }
 	  mix_unlock("mail",lockptr);
 
@@ -524,6 +536,7 @@ printf("final hop\n"); /* debug */
 	  byteptr += 8;
 	  if(!check_packetID(packetID)) return(-1);
 	  crypt_in_buffer(innerkey,iv,body,0); /* decrypt back into the buffer */
+	  add_to_random(body->message,body->length);
 
 	  /* get destination list */
 	  byteptr = body->message;
@@ -562,12 +575,13 @@ printf("final hop\n"); /* debug */
 	  strcpy(line,"mail");
 	  if((fptr = tempfile(line))==NULL) return(-1);
 
-	  /* put the destination list on the first line */
+          fwrite(final_hop,1,strlen(final_hop),fptr);
+	  /* put the destination list in the file, one per line */
 	  for(i=0;i<numdest;i++) {
 	    fwrite(destination[i],1,strlen(destination[i]),fptr);
-	    fwrite(" ",1,1,fptr);
+	    fwrite("\n",1,1,fptr);
 	  }
-	  fwrite("\n",1,1,fptr);
+	  fwrite("END\n",1,4,fptr);
 
 	  /* insert disclaimers */
 	  b3 = new_buffer();
@@ -605,6 +619,7 @@ printf("packet final hop\n"); /* debug */
 	  byteptr += 8;
 	  if(!check_packetID(packetID)) return(-1);
 	  crypt_in_buffer(innerkey,iv,body,0); /* decrypt back into the buffer */
+	  add_to_random(body->message,body->length);
 
 	  mix_lock("pac",&lockptr);
 	  merge_packets(body, messageID, packet, numpackets);
@@ -642,9 +657,10 @@ int read_remailer_list(FILE *ptr, REMAILER *list)
 }
 
 /* send packet just generated */
-int send_new_packet(headbuff,bodybuff, address,outfile)
+int send_new_packet(headbuff,bodybuff, address,outfile, client)
 BUFFER	**headbuff, *bodybuff;
 char	*outfile,*address;
+int	client;
 {
 	int	i,j,k,n,len;
 	BUFFER	*outbuffer,*tmpbuffer;
@@ -664,12 +680,20 @@ char	*outfile,*address;
 
 	/* build output file */
 	outbuffer = new_buffer();
-	add_to_buffer(outbuffer, "\n::\n",4);
+	/* If either client being mailed directly, or not client then add a To: line */
+	if(strlen(outfile)<=1 || !client) {
+ 	  add_to_buffer(outbuffer, "To: ", strlen("To: "));
+	  add_to_buffer(outbuffer, address, strlen(address));
+	}
+	add_to_buffer(outbuffer, "\n\n::\n",strlen("\n\n::\n"));
 	add_to_buffer(outbuffer, remailer_type, strlen(remailer_type));
 	add_to_buffer(outbuffer, VERSION, strlen(VERSION));
 	add_to_buffer(outbuffer, two_n, strlen(two_n));
 	add_to_buffer(outbuffer, begin_remailer,strlen (begin_remailer));
 	sprintf(line,"%d\n",tmpbuffer->length); /* Write out length */
+#ifdef DEBUGGING
+/* debug */	fprintf(stdout,"%s\n",outbuffer->message); /* Write out length */
+#endif
 	add_to_buffer(outbuffer, line,strlen(line));
 
 	/* Ok, now make a MD5 checksum and prepend it. */
@@ -698,10 +722,11 @@ char	*outfile,*address;
         add_to_buffer (outbuffer, end_remailer, strlen (end_remailer));
 
 	if(strlen(outfile)<=1) {
-	  strcpy(temp,SENDMAIL);
-	  strcat(temp,address);
-          fptr = popen(temp,"w");
-          if(fptr == NULL) return(-1);
+          fptr = popen(SENDMAIL,"w");
+          if(fptr == NULL) {
+            fprintf(stderr,"Error: Unable to open pipe to %s\n",SENDMAIL);
+            return(-1);
+          }
           fwrite(outbuffer->message,sizeof(char),outbuffer->length,fptr);
 	  fclose(fptr);
 	} else if(strstr(outfile,"stdout")) {
@@ -709,7 +734,10 @@ char	*outfile,*address;
 	} else {
 	  fptr = fopen(outfile,"w");
           if(fptr == NULL) return(-1);
-	  fprintf(fptr,"%s\n",address);
+	  if(!client) { /* then we are a remailer and should add some headers */
+	    fprintf(fptr,"%s",intermed_hop);
+	    fprintf(fptr,"%s\nEND\n",address);
+	  }
           fwrite(outbuffer->message,sizeof(char),outbuffer->length,fptr);
 	  flock(fileno(fptr),LOCK_UN); /* possible locked before calling */
 	  fclose(fptr);
@@ -741,11 +769,6 @@ REMAILER 	*remailer_list;
 	R_RSA_PUBLIC_KEY  pubKey,*keyPtr[5];
 	unsigned int	numPubKeys,keylen;
 
-	/* Grab some random numbers */
-	while(fread(line,1,256,in_ptr)>0) {
-	   add_to_random(line,256);
-	}
-	rewind(in_ptr);
 
 	/* compress the file */
 	/* ?? Compression would go here ?? */
@@ -773,6 +796,7 @@ REMAILER 	*remailer_list;
 	   add_to_buffer(message,line,i);
 	}
 	fclose(in_ptr);
+	add_to_random(message->message, message->length);
 	numpackets = message->length / PACKETSIZE;
 	if(message->length % PACKETSIZE != 0) numpackets++;
 
@@ -824,8 +848,19 @@ REMAILER 	*remailer_list;
 	  }
 
 	  for(hop = remailer_chain[0]; hop >= 1; hop--){
+	    /* Check to see if this remailer should be chosen randomly */
+	    if(remailer_chain[hop] <= 0) {
+	      do {
+	        tmpbyte = our_randombyte();
+	      } while (tmpbyte >= num_remailers);
+	      tmpbyte ++; /* range of 1 to num_remailers */
+	      remailer_chain[hop] = -1 * tmpbyte;
+	      /* The last hop has to be the same for all packets */
+	      if(hop == remailer_chain[0])
+	           remailer_chain[hop] = abs(remailer_chain[hop]);
+	    }
 	    /* Get RSA key for remailer */
-	    if(get_rsa_key(remailer_list[remailer_chain[hop]].key_ID,
+	    if(get_rsa_key(remailer_list[abs(remailer_chain[hop])].key_ID,
 	    	&pubKey)!=0) {
 	       fprintf(stderr,"Can't get public key!\n");
 	       return(-1);
@@ -880,7 +915,7 @@ REMAILER 	*remailer_list;
 	      for(i=0;i<19;i++) {
 	         add_to_buffer(headbuff[hop],ivarray[i],8);
 	      }
-	      add_to_buffer(headbuff[hop],remailer_list[remailer_chain[hop+1]].name,80);
+	      add_to_buffer(headbuff[hop],remailer_list[abs(remailer_chain[hop+1])].name,80);
 	    }
 
 
@@ -915,7 +950,7 @@ REMAILER 	*remailer_list;
 	    clear_buffer(headbuff[hop]);
 
 	    /* Prepend RSA key ID */
-	    add_to_buffer(headbuff[hop],remailer_list[remailer_chain[hop]].key_ID,16);
+	    add_to_buffer(headbuff[hop],remailer_list[abs(remailer_chain[hop])].key_ID,16);
 	    /* prepend keys and IV to header */
 	    tmpbyte = keylen;
 	    add_to_buffer(headbuff[hop],&tmpbyte,1);
@@ -944,9 +979,9 @@ REMAILER 	*remailer_list;
 	  } /* hop loop for a given packet */
 	  if(strlen(outfile) > 1 && !strstr(outfile,"stdout") && numpackets > 1) {
 	    sprintf(line,"%s.%d",outfile,packet);
-	    send_new_packet(headbuff,bodybuff, remailer_list[remailer_chain[1]].name,"line");
+	    send_new_packet(headbuff,bodybuff, remailer_list[abs(remailer_chain[1])].name,line,1);
 	  } else {
-	    send_new_packet(headbuff,bodybuff, remailer_list[remailer_chain[1]].name,outfile);
+	    send_new_packet(headbuff,bodybuff, remailer_list[abs(remailer_chain[1])].name,outfile,1);
 	  }
 	} /* end loop proccessing packets */
 	free_buffer(bodybuff);
@@ -970,7 +1005,8 @@ void get_chain(REMAILER *remailer_list, int num_remailers, int *chain)
 	chain[0]=0;
 	while(line[0] >='0' && line[0]<='9') {
 	  j = atoi(line);
-	  if (j>=1 && j<=num_remailers && chain[0] < 20) {
+	  /* remailer 0 means choose randomly */
+	  if (j>=0 && j<=num_remailers && chain[0] < 20) {
 	    chain[0]++;
 	    chain[chain[0]] = j; /* because chain[0] holds the number of elements */
 	  } else {
@@ -1008,6 +1044,13 @@ int	chain_2(int argc, char *argv[])
 	for(i=1; i<argc; i++) {
 	  if(strstr(argv[i],"-c")) {
 	    /* nop */
+	  } else if(strstr(argv[i],"-h") || strstr(argv[i],"--help")) {
+	    /* Print help and exit */
+	    printf("Mixmaster (C) Copyright Lance M. Cottrell 1995\n");
+	    printf("Released under the GNU public license.\n");
+	    printf("Client Mode command line arguments:\n");
+	    printf("mixmaster [-c][-f][filename][-o outfile][-to who@where][-s \"subject\"][-l 1 2 3 4]\n");
+	    exit(-1);
 	  } else if(strstr(argv[i],"-f")) {
 	    /* set filter mode */
 	    filter = 1;
@@ -1035,7 +1078,8 @@ int	chain_2(int argc, char *argv[])
 	    while(i < argc && j < 20){
 	      j++;
 	      remailer_chain[j] = atoi(argv[i]);
-	      if(remailer_chain[j] <=0 || remailer_chain[j] > num_remailers){
+	      /* remailer = 0 means random remailer */
+	      if(remailer_chain[j] <0 || remailer_chain[j] > num_remailers){
 		fprintf(stderr,"Invalid remailer number!\n");
 		exit(-1);
 	      }
